/* FILE: meshvalue.h                 -*-Mode: c++-*-
 *
 * Oxs_MeshValue templated class, intended for use with the
 * Oxs_Mesh family.  For dependency reasons, the class definitions
 * are in the meshvalue_defn.h header file, which may be used by
 * classes that need definitions but not implementation details.
 *
 * Note: To maintain a clean dependency hierarchy and prevent cyclic
 *   dependencies, this class does not maintain a lock on the mesh
 *   used to construct it.  Only the size of the mesh is retained,
 *   although a client may call Oxs_MeshValue<T>::CheckMesh() to
 *   check that the size is compatible with a given mesh.  If
 *   Oxs_MeshValue becomes more complex in the future, it may be
 *   necessary to add a pointer to the mesh that construction was
 *   based upon, in which case it will be the responsibility of the
 *   holder and/or user of the Oxs_MeshValue object to insure the
 *   correctness of that pointer.
 */

#ifndef _OXS_MESHVALUE
#define _OXS_MESHVALUE

#include "meshvalue_defn.h"
#include "mesh.h"

/* End includes */

////////////////////////////////////////////////////////////////////////
// Include all implementations here.  This is required by compiler/linkers
// using what gcc calls the "Borland model," as opposed to the "Cfront
// model."  The latter is distinguished by the use of template
// repositories.  With the Cfront model, template definitions can be
// put in a file (*.cc) separate from the header file (*.h).  Putting
// the definitions in the header file along with declarations slows down
// compilation, because the definitions are re-compiled in each including
// file, but appears to be more portable.

template<class T> BOOL
Oxs_MeshValue<T>::CheckMesh(const Oxs_Mesh* mesh) const
{ // Returns true if size is compatible with mesh
  if(mesh==NULL) return 0;
  return (mesh->Size() == size);
}

template<class T>
const T& Oxs_MeshValue<T>::operator[](UINT4m index) const
{
#ifndef NDEBUG
  if(index>=size) {
    char buf[512];
    Oc_Snprintf(buf,sizeof(buf),
                "const T& Oxs_MeshValue<T>::operator[] const: "
                "Array out-of-bounds; index=%u, size=%u",
                index,size);
    OXS_THROW(Oxs_BadIndex,buf);
  }
#endif
  return arr[index];
}

template<class T>
T& Oxs_MeshValue<T>::operator[](UINT4m index)
{
#ifndef NDEBUG
  if(index>=size) {
    char buf[512];
    Oc_Snprintf(buf,sizeof(buf),
                "T& Oxs_MeshValue<T>::operator[]: "
                "Array out-of-bounds; index=%u, size=%u",
                index,size);
    OXS_THROW(Oxs_BadIndex,buf);
  }
#endif
  return arr[index];
}

template<class T>
void Oxs_MeshValue<T>::Free()
{
  // Free value array
  if(arr!=NULL) { delete[] arr; arr=NULL; }
  if(size!=0) {
    size=0;
    SetPieceData();
  }
}

////////////////////////////////////////////////////////////////////////
/// CONSTRUCTORS

#if OOMMF_THREADS
template<class T>
void Oxs_MeshValue<T>::SetPieceData()
{ // This routine assumes member values size, max_number_of_threads
  // and minimum_thread_size are already set.
  int trycount = size/minimum_thread_size;
  if(trycount>max_number_of_threads) trycount=max_number_of_threads;
  if(trycount<1)                     trycount=1;
  piecedata.resize(trycount);
  int piecesize = (size+trycount-1)/trycount;
  size_t index = 0;
  for(int i=0;i<trycount;++i) {
    piecedata[i].istart = index;
    index += piecesize;
    if(index>size) index=size;
    piecedata[i].istop = index;
  }
}
#endif // OOMMF_THREADS

template<class T>
Oxs_MeshValue<T>::Oxs_MeshValue()
  : size(0), arr(NULL)
#if OOMMF_THREADS
    , max_number_of_threads(4),     // Hard code for now
    minimum_thread_size(1000)   // Hard code for now
#endif
{
  SetPieceData();
}

template<class T>
Oxs_MeshValue<T>::Oxs_MeshValue(const Oxs_Mesh* mesh)
  : size(0), arr(NULL)
#if OOMMF_THREADS
    , max_number_of_threads(4),     // Hard code for now
    minimum_thread_size(1000)   // Hard code for now
#endif
{
  if(mesh!=NULL) {
    // Allocate value array
    size=mesh->Size();
    arr = new T[size];
    if(arr==NULL) {
      char errbuf[128];
      Oc_Snprintf(errbuf,sizeof(errbuf),
                  "Unable to allocate memory for %lu objects of size %u",
                  (unsigned long)size,(unsigned)sizeof(T));
      OXS_THROW(Oxs_NoMem,errbuf);
    }
    SetPieceData();
  }
}

template<class T>
Oxs_MeshValue<T>::~Oxs_MeshValue()
{
  Free();
}

template<class T>
Oxs_MeshValue<T>::Oxs_MeshValue(const Oxs_MeshValue<T> &other)
{ // Copy constructor.
  size=other.size;
  if(size>0) {
    arr = new T[size];
    if(arr==NULL) {
      char errbuf[128];
      Oc_Snprintf(errbuf,sizeof(errbuf),
                  "Unable to allocate memory for %lu objects of size %u",
                  (unsigned long)size,(unsigned)sizeof(T));
      OXS_THROW(Oxs_NoMem,errbuf);
    }
    memcpy(arr,other.arr,static_cast<size_t>(size)*sizeof(T));
    /// Should we thread this?
  } else {
    arr = NULL;
  }
#if OOMMF_THREADS
  max_number_of_threads = other.max_number_of_threads;
  minimum_thread_size = other.minimum_thread_size;
  SetPieceData();
#endif
}

template<class T>
void Oxs_MeshValue<T>::Swap(Oxs_MeshValue<T>& other)
{ // Note: This works okay also if this==&other
  UINT4m tsize = size;  size = other.size;  other.size = tsize;
  T*     tarr  = arr;   arr  = other.arr;   other.arr  = tarr;
#if OOMMF_THREADS
  int tnumber = max_number_of_threads;
  max_number_of_threads = other.max_number_of_threads;
  other.max_number_of_threads = tnumber;

  UINT4m tminsize = minimum_thread_size;
  minimum_thread_size = other.minimum_thread_size;
  other.minimum_thread_size = tminsize;

  piecedata.swap(other.piecedata);
#endif
}

template<class T>
void Oxs_MeshValue<T>::AdjustSize(const Oxs_Mesh* newmesh)
{
  if(newmesh == NULL) {
    Free();
  } else {
    if(newmesh->Size() != size) {
      // Allocate new value array
      if(arr!=NULL) { delete[] arr; arr=NULL; }
      size=newmesh->Size();
      if(size>0) {
        arr = new T[size];
        if(arr==NULL) {
          char errbuf[128];
          Oc_Snprintf(errbuf,sizeof(errbuf),
                      "Unable to allocate memory for %lu objects of size %u",
                      (unsigned long)size,(unsigned)sizeof(T));
          OXS_THROW(Oxs_NoMem,errbuf);
        }
      }
      SetPieceData();
    }
  }
}

template<class T> Oxs_MeshValue<T>&
Oxs_MeshValue<T>::operator=(const Oxs_MeshValue<T>& other)
{
  if(this == &other) return *this;  // Nothing to do.

  if(size!=other.size) {
    Free();
    size=other.size;
    if(size>0) {
      arr = new T[size];
      if(arr==NULL) {
        char errbuf[128];
        Oc_Snprintf(errbuf,sizeof(errbuf),
                    "Unable to allocate memory for %lu objects of size %u",
                    (unsigned long)size,(unsigned)sizeof(T));
        OXS_THROW(Oxs_NoMem,errbuf);
      }
    } else {
      arr = NULL;  // Safety
    }
    // Question: Should we copy max_number_of_threads and minimum_thread_size
    // from other, or keep current settings?  As a first pass, we'll keep
    // the current settings.  Note, though, that the copy constuctor
    // *does* use the "other" values.
    SetPieceData();
  }

  memcpy(arr,other.arr,static_cast<size_t>(size)*sizeof(T));
  /// Should we thread this?

  return *this;
}

////////////////////////////////////////////////////////////////////////
// Array operators.  Each has separate single-threaded and
// multi-threaded implementations.

#if !OOMMF_THREADS  // SINGLE THREADED IMPLEMENTATIONS /////////////////

template<class T> Oxs_MeshValue<T>&
Oxs_MeshValue<T>::operator+=(const Oxs_MeshValue<T>& other)
{
  if(size!=other.size) {
    OXS_THROW(Oxs_BadIndex,
             "Size mismatch in Oxs_MeshValue<T>::operator+=");
  }
  for(UINT4m i=0;i<size;i++) arr[i]+=other.arr[i];
  return *this;
}

template<class T> Oxs_MeshValue<T>&
Oxs_MeshValue<T>::operator-=(const Oxs_MeshValue<T>& other)
{
  if(size!=other.size) {
    OXS_THROW(Oxs_BadIndex,
             "Size mismatch in Oxs_MeshValue<T>::operator-=");
  }
  for(UINT4m i=0;i<size;i++) arr[i]-=other.arr[i];
  return *this;
}

template<class T> Oxs_MeshValue<T>&
Oxs_MeshValue<T>::operator*=(REAL8m mult)
{
  for(UINT4m i=0;i<size;i++) arr[i]*=mult;
  return *this;
}


template<class T> Oxs_MeshValue<T>&
Oxs_MeshValue<T>::operator=(const T &value)
{
  for(UINT4m i=0;i<size;i++) arr[i]=value;
  return *this;
}

template<class T> Oxs_MeshValue<T>&
Oxs_MeshValue<T>::Accumulate
(REAL8m mult,
 const Oxs_MeshValue<T> &other)
{
  if(size!=other.size) {
    OXS_THROW(Oxs_BadIndex,
             "Size mismatch in Oxs_MeshValue<T>::Accumulate");
  }
  for(UINT4m i=0;i<size;i++) arr[i]+=mult*other.arr[i];
  return *this;
}

#else // OOMMF_THREADS // MULTI-THREADED IMPLEMENTATIONS ///////////////

template<class T>
class _Oxs_MeshValuePlusEqual : public Oxs_ThreadRunObj {
public:
  void Cmd(int threadnumber, void* data);
};
template<class T>
void _Oxs_MeshValuePlusEqual<T>::Cmd(int,void* data)
{
  _Oxs_MeshValueThreadData<T>& thread_data 
    = *static_cast<_Oxs_MeshValueThreadData<T>*>(data);
  const INT4m istart = static_cast<INT4m>(thread_data.istart);
  const INT4m istop  = static_cast<INT4m>(thread_data.istop);
  // Note: int's run faster than unsigned on some platforms,
  //       notably the DEC Alpha.  We assume overflow of
  //       INT4m is *not* an issue!
  T* const A = thread_data.A;
  const T* const B = thread_data.B;
  for(INT4m i=istart;i<istop;++i) {
    A[i] += B[i];
  }
}

template<class T> Oxs_MeshValue<T>&
Oxs_MeshValue<T>::operator+=(const Oxs_MeshValue<T>& other)
{
  if(size!=other.size) {
    OXS_THROW(Oxs_BadIndex,
             "Size mismatch in Oxs_MeshValue<T>::operator+=");
  }
  _Oxs_MeshValuePlusEqual<T>  plusequal;
  const int thread_count = piecedata.size();
  for(int piece=1;piece<thread_count;++piece) {
    piecedata[piece].A = arr;
    piecedata[piece].B = other.arr;
    threadtree.Launch(plusequal,&piecedata[piece]);
  }
  piecedata[0].A = arr;
  piecedata[0].B = other.arr;
  plusequal.Cmd(0,&piecedata[0]);
  threadtree.Join();
  return *this;
}

template<class T>
class _Oxs_MeshValueMinusEqual : public Oxs_ThreadRunObj {
public:
  void Cmd(int threadnumber, void* data);
};
template<class T>
void _Oxs_MeshValueMinusEqual<T>::Cmd(int,void* data)
{
  _Oxs_MeshValueThreadData<T>& thread_data 
    = *static_cast<_Oxs_MeshValueThreadData<T>*>(data);
  const INT4m istart = static_cast<INT4m>(thread_data.istart);
  const INT4m istop  = static_cast<INT4m>(thread_data.istop);
  T* const A = thread_data.A;
  const T* const B = thread_data.B;
  for(INT4m i=istart;i<istop;++i) A[i] -= B[i];
}

template<class T> Oxs_MeshValue<T>&
Oxs_MeshValue<T>::operator-=(const Oxs_MeshValue<T>& other)
{
  if(size!=other.size) {
    OXS_THROW(Oxs_BadIndex,
             "Size mismatch in Oxs_MeshValue<T>::operator-=");
  }
  _Oxs_MeshValueMinusEqual<T> minusequal;
  const int thread_count = piecedata.size();
  for(int piece=1;piece<thread_count;++piece) {
    piecedata[piece].A = arr;
    piecedata[piece].B = other.arr;
    threadtree.Launch(minusequal,&piecedata[piece]);
  }
  piecedata[0].A = arr;
  piecedata[0].B = other.arr;
  minusequal.Cmd(0,&piecedata[0]);
  threadtree.Join();
  return *this;
}

template<class T>
class _Oxs_MeshValueTimesEqual : public Oxs_ThreadRunObj {
public:
  void Cmd(int threadnumber, void* data);
};
template<class T>
void _Oxs_MeshValueTimesEqual<T>::Cmd(int,void* data)
{
  _Oxs_MeshValueThreadData<T>& thread_data 
    = *static_cast<_Oxs_MeshValueThreadData<T>*>(data);
  const INT4m istart = static_cast<INT4m>(thread_data.istart);
  const INT4m istop  = static_cast<INT4m>(thread_data.istop);
  T* const A = thread_data.A;
  const REAL8m mult = thread_data.mult;
  for(INT4m i=istart;i<istop;++i) A[i] *= mult;
}

template<class T> Oxs_MeshValue<T>&
Oxs_MeshValue<T>::operator*=(REAL8m mult)
{
  _Oxs_MeshValueTimesEqual<T> timesequal;
  const int thread_count = piecedata.size();
  for(int piece=1;piece<thread_count;++piece) {
    piecedata[piece].A    = arr;
    piecedata[piece].mult = mult;
    threadtree.Launch(timesequal,&piecedata[piece]);
  }
  piecedata[0].A    = arr;
  piecedata[0].mult = mult;
  timesequal.Cmd(0,&piecedata[0]);
  threadtree.Join();
  return *this;
}


template<class T>
class _Oxs_MeshValueValueEqual : public Oxs_ThreadRunObj {
public:
  void Cmd(int threadnumber, void* data);
};
template<class T>
void _Oxs_MeshValueValueEqual<T>::Cmd(int,void* data)
{
  _Oxs_MeshValueThreadData<T>& thread_data 
    = *static_cast<_Oxs_MeshValueThreadData<T>*>(data);
  const INT4m istart = static_cast<INT4m>(thread_data.istart);
  const INT4m istop  = static_cast<INT4m>(thread_data.istop);
  T* const A = thread_data.A;
  const T& value = thread_data.value;
  for(INT4m i=istart;i<istop;++i) A[i] = value;
}

template<class T> Oxs_MeshValue<T>&
Oxs_MeshValue<T>::operator=(const T &value)
{
  _Oxs_MeshValueValueEqual<T> valueequal;
  const int thread_count = piecedata.size();
  for(int piece=1;piece<thread_count;++piece) {
    piecedata[piece].A     = arr;
    piecedata[piece].value = value;
    threadtree.Launch(valueequal,&piecedata[piece]);
  }
  piecedata[0].A     = arr;
  piecedata[0].value = value;
  valueequal.Cmd(0,&piecedata[0]);
  threadtree.Join();
  return *this;
}

template<class T>
class _Oxs_MeshValueAccumulate : public Oxs_ThreadRunObj {
public:
  void Cmd(int threadnumber, void* data);
};
template<class T>
void _Oxs_MeshValueAccumulate<T>::Cmd(int,void* data)
{
  _Oxs_MeshValueThreadData<T>& thread_data 
    = *static_cast<_Oxs_MeshValueThreadData<T>*>(data);
  const INT4m istart = static_cast<INT4m>(thread_data.istart);
  const INT4m istop  = static_cast<INT4m>(thread_data.istop);
  T* const A = thread_data.A;
  const T* const B = thread_data.B;
  const REAL8m mult = thread_data.mult;
  for(INT4m i=istart;i<istop;++i) A[i] += mult*B[i];
}

template<class T> Oxs_MeshValue<T>&
Oxs_MeshValue<T>::Accumulate
(REAL8m mult,
 const Oxs_MeshValue<T> &other)
{
  if(size!=other.size) {
    OXS_THROW(Oxs_BadIndex,
             "Size mismatch in Oxs_MeshValue<T>::Accumulate");
  }
  _Oxs_MeshValueAccumulate<T> accumulate;
  const int thread_count = piecedata.size();
  for(int piece=1;piece<thread_count;++piece) {
    piecedata[piece].A    = arr;
    piecedata[piece].B    = other.arr;
    piecedata[piece].mult = mult;
    threadtree.Launch(accumulate,&piecedata[piece]);
  }
  piecedata[0].A    = arr;
  piecedata[0].B    = other.arr;
  piecedata[0].mult = mult;
  accumulate.Cmd(0,&piecedata[0]);
  threadtree.Join();
  return *this;
}

#endif // OOMMF_THREADS ////////////////////////////////////////////////

#endif // _OXS_MESHVALUE
