/* FILE: meshvalue_defn.h                 -*-Mode: c++-*-
 *
 * Oxs_MeshValue templated class, intended for use with the
 * Oxs_Mesh family.  This file contains the class definitions
 * but not the implementations, which are primarily in the file
 * meshvalue.h.  The implementations are in a header file due
 * to some (historic?) details involving template instantiation.
 * These definitions are broken out of that file due to dependency
 * considerations involving Oxs_Ext, Oxs_Output, and this class.
 *
 * Note: To maintain a clean dependency hierarchy and prevent cyclic
 *   dependencies, this class does not maintain a lock on the mesh
 *   used to construct it.  Only the size of the mesh is retained,
 *   although a client may call Oxs_MeshValue<T>::CheckMesh() to
 *   check that the size is compatible with a given mesh.  If
 *   Oxs_MeshValue becomes more complex in the future, it may be
 *   necessary to add a pointer to the mesh that construction was
 *   based upon, in which case it will be the responsibility of the
 *   holder and/or user of the Oxs_MeshValue object to insure the
 *   correctness of that pointer.
 */

#ifndef _OXS_MESHVALUE_DEFN
#define _OXS_MESHVALUE_DEFN

#include <string.h>  // For memcpy

#if OOMMF_THREADS
# include <vector>
#endif

#include "oxsexcept.h"
#include "oxsthread.h"

/* End includes */

class Oxs_Mesh;  // Forward declaration

#if OOMMF_THREADS
// Support structure for multi-threaded version
template<class T>
struct _Oxs_MeshValueThreadData {   // Each thread gets one of these
  T* A;          // structs, which tells it the starting and ending
  T* B;          // index for that thread (istart and istop).  The
  T value;       // T array pointers A, B, and values "value" and "mult"
  REAL8m mult;   // are filled in as needed when threads are actived.
  UINT4m istart;
  UINT4m istop;     
  _Oxs_MeshValueThreadData() : A(0), B(0), mult(0.0), istart(0), istop(0) {}
  /// NB: It is assumed that T has a default initializer.
};

#endif // OOMMF_THREADS

template<class T> class Oxs_MeshValue {
 private:
  UINT4m size;
  T* arr;
  void Free();

#if OOMMF_THREADS
  // For multi-threaded version
  int max_number_of_threads;
  UINT4m minimum_thread_size; // If ::size is smaller than this,
  /// then run a single thread instead of splitting.
  vector< _Oxs_MeshValueThreadData<T> > piecedata;
  void SetPieceData();
  Oxs_ThreadTree threadtree;
#else
  void SetPieceData() {}  // For convenience
#endif // OOMMF_THREADS

 public:
  Oxs_MeshValue();
  Oxs_MeshValue(const Oxs_Mesh* mesh);
  ~Oxs_MeshValue();

  Oxs_MeshValue(const Oxs_MeshValue<T> &other);  // Copy constructor
  Oxs_MeshValue<T>& operator=(const Oxs_MeshValue<T> &other);
  Oxs_MeshValue<T>& operator+=(const Oxs_MeshValue<T> &other);
  Oxs_MeshValue<T>& operator-=(const Oxs_MeshValue<T> &other);
  Oxs_MeshValue<T>& operator*=(REAL8m mult);

  Oxs_MeshValue<T>& operator=(const T &value); // Fills out
  /// *this with size copies of value

  Oxs_MeshValue<T>& Accumulate(REAL8m mult,
                               const Oxs_MeshValue<T> &other);

  void Swap(Oxs_MeshValue<T>& other);

  void Release() { Free(); }  // Frees arr
  void AdjustSize(const Oxs_Mesh* newmesh);  // Compares current size
  /// to size of newmesh.  If different, deletes current arr and
  /// allocates new one compatible with newmesh.

  BOOL CheckMesh(const Oxs_Mesh* mesh) const;
  /// Returns true if current arr size is compatible with mesh.

  inline UINT4m Size() const { return size; }
  inline const T& operator[](UINT4m index) const;
  inline T& operator[](UINT4m index);
};

#endif // _OXS_MESHVALUE_DEFN
