/* FILE: output.h                 -*-Mode: c++-*-
 *
 * Output classes.  Each class providing output should embed an
 * instance of an output class for each output provided.
 *
 * NOTE: This code looks pretty fragile, because there is a lot
 *  of coupling between the output classes and their owner class.
 *  Maybe we can figure out a cleaner, more robust solution in
 *  the future.
 */

#ifndef _OXS_OUTPUT
#define _OXS_OUTPUT

#include "output_defn.h"

#include "ext.h"
#include "simstate.h"

/* End includes */

////////////////////////////////////////////////////////////////////////
// Include definitions below.  This is required by compiler/linkers
// using what gcc calls the "Borland model," as opposed to the "Cfront
// model."  The latter is distinguished by the use of template
// repositories.  With the Cfront model, template definitions can be
// put in a file (*.cc) separate from the header file (*.h).  Putting
// the definitions in the header file along with declarations slows down
// compilation, because the definitions are re-compiled in each including
// file, but appears to be more portable.

//////////////////////////////////////////////////////////////
// Oxs_ScalarOutput<T> definitions
template<class T>
int Oxs_ScalarOutput<T>::SetOutputFormat
(const String& format)
{
  if(format.length()==0) return 1;
  output_format = format;
  return 0;
}

template<class T>
int Oxs_ScalarOutput<T>::Output
(const Oxs_SimState* state,   // State to get output for
 Tcl_Interp* interp,          // Tcl interpreter
 int argc,CONST84 char** /* argv */)  // Args
{
  Tcl_ResetResult(interp);
  if(argc!=0) {
    Tcl_AppendResult(interp, "wrong # of args: should be none",
                     (char *) NULL);
    return TCL_ERROR;
  }
  if(state==NULL || state->Id()==0) {
    Tcl_AppendResult(interp,"Invalid state passed to output object",
                     (char *)NULL);
    return TCL_ERROR;
  }

  if(!IsCacheSupported() || cache.state_id!=state->Id()) {
    // Fill cache with data
    ForceCacheIncrement(1);
    cache.state_id=0; // Safety
    (owner->*getscalar)(*state);
    ForceCacheIncrement(-1);

    // Check that cache got filled
    if(cache.state_id!=state->Id()) {
      OXS_THROW(Oxs_ProgramLogicError,
               "Output cache not filled by registered getscalar"
               " call (as required).");
    }
  }

  char buf[100];
  Oc_Snprintf(buf,sizeof(buf),
              output_format.c_str(),cache.value);

  Tcl_ResetResult(interp); // Protection against badly behaved Oxs_Ext's.
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}


//////////////////////////////////////////////////////////////
// Oxs_VectorFieldOutput<T> definitions
template<class T>
int Oxs_VectorFieldOutput<T>::SetOutputFormat
(const String& format)
{
  if(format.length()==0) return 1;
  Oxs_SplitList arglist;
  arglist.Split(format.c_str());
  if(arglist.Count()!=2) {
    OXS_THROW(Oxs_BadParameter,
              "Wrong number of arguments in format string import to "
              "Oxs_VectorFieldOutput<T>::SetOutputFormat(const char*)");
  }
  output_format_datatype = arglist[0];
  output_format_precision = arglist[1];
  return 0;
}

template<class T>
String Oxs_VectorFieldOutput<T>::GetOutputFormat() const
{
  vector<String> sarr;
  sarr.push_back(output_format_datatype);
  sarr.push_back(output_format_precision);
  return Oxs_MergeList(&sarr);
}

template<class T>
int Oxs_VectorFieldOutput<T>::SetOutputMeshtype
(const String& meshtype)
{
  if(meshtype.length()==0) return 1;
  output_format_meshtype = meshtype;
  return 0;
}

template<class T>
String Oxs_VectorFieldOutput<T>::GetOutputMeshtype() const
{
  return output_format_meshtype;
}

template<class T>
int Oxs_VectorFieldOutput<T>::UpdateCache
(const Oxs_SimState* state)
{ // Returns 1 on success, 0 if caching not supported or not enabled.
  // Throws an exception if an error occurs.

  if(!IsCacheSupported() || GetCacheRequestCount()<1) return 0;

  if(cache.state_id!=state->Id()) {
    ForceCacheIncrement(1);
    cache.state_id=0; // Safety
    cache.value.AdjustSize(state->mesh);
    (owner->*getvectorfield)(*state);
    ForceCacheIncrement(-1);

    // Check that cache got filled
    if(cache.state_id!=state->Id()) {
      OXS_THROW(Oxs_ProgramLogicError,
               "Output cache not filled by registered getvectorfield"
               " call (as required).");
    }
  }

  return 1;
}


template<class T>
int Oxs_VectorFieldOutput<T>::Output
(const Oxs_SimState* state, // State to get output for
 Tcl_Interp* interp,        // Tcl interpreter
 int argc,CONST84 char** argv)      // Args
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Tcl_AppendResult(interp, "wrong # of args: should be 1: filename",
                     (char *) NULL);
    return TCL_ERROR;
  }
  if(state==NULL || state->Id()==0) {
    Tcl_AppendResult(interp,"Invalid state passed to output object",
                     (char *)NULL);
    return TCL_ERROR;
  }

  const char* filename = argv[0];

  if(!IsCacheSupported() || cache.state_id!=state->Id()) {
    // Fill cache with vector field
    ForceCacheIncrement(1);
    cache.state_id=0; // Safety
    cache.value.AdjustSize(state->mesh);
    (owner->*getvectorfield)(*state);
    ForceCacheIncrement(-1);

    // Check that cache got filled
    if(cache.state_id!=state->Id()) {
      OXS_THROW(Oxs_ProgramLogicError,
               "Output cache not filled by registered getvectorfield"
               " call (as required).");
    }
  }

  // Write cache to file
  char description[4096];
  Oc_Snprintf(description,sizeof(description),
              "Oxs %.80s output\n"
              " MIF source file: %.256s\n"
              " Iteration: %u, State id: %u\n"
              " Stage: %u, Stage iteration: %u\n"
              " Stage simulation time: %g s\n"
              " Total simulation time: %g s\n",
              OutputType().c_str(),
              GetProblemName(),
              state->iteration_count,state->Id(),
              state->stage_number,state->stage_iteration_count,
              state->stage_elapsed_time,
              state->stage_start_time + state->stage_elapsed_time);
  int errorcode = TCL_OK;
  try {
    state->mesh->WriteOvf(filename,
                          GetOutputWriteheaders(),
                          LongName().c_str(),
                          description,
                          OutputUnits().c_str(),
                          output_format_meshtype.c_str(),
                          output_format_datatype.c_str(),
                          output_format_precision.c_str(),
                          &cache.value,NULL);
  } catch (Oxs_Ext::Error& err) {
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
                "Error writing Ovf file %s, state id %u: ",
                filename,state->Id());
    Tcl_ResetResult(interp); // Protection against badly behaved Oxs_Ext's.
    Tcl_AppendResult(interp,buf,err.c_str(),(char *)NULL);
    errorcode = TCL_ERROR;
  } catch (...) {
    // Unexpected exception; Clean-up and rethrow.
    if(GetCacheRequestCount()<1) {
      cache.state_id=0;
      cache.value.Release();
    }
    throw;
  }

  // Release buffer if caching is disabled
  if(GetCacheRequestCount()<1) {
    cache.state_id=0;
    cache.value.Release();
  }

  // On success, return value is the filename.
  if(errorcode==TCL_OK) {
    Tcl_ResetResult(interp); // Protection against badly behaved Oxs_Ext's.
    Tcl_AppendResult(interp,filename,(char *)NULL);
  }
  return errorcode;
}


#endif // _OXS_OUTPUT
