/* FILE: oxsthread.cc              -*-Mode: c++-*-
 *
 * OXS thread support.
 *
 * These classes are only compiled if the OXS_DO_THREADS macro
 * is defined.
 */

#if OOMMF_THREADS

#include "oxsexcept.h"
#include "oxsthread.h"

/* End includes */

// Notes on Tcl_MutexLock, Tcl_ConditionNotify, and Tcl_ConditionWait:
//  1) A lock should be held on a mutex before issuing a
//     notify or performing a wait.
//  2) When a wait is issued, the mutex lock is released.
//     When the thread awakens, it automatically reclaims
//     the mutex lock.
//  3) After issuing a notify, the code needs to explicitly
//     release the mutex lock.  Threads waiting on the
//     notify condition will not begin executing until
//     they can grab the mutex (as explained in 2).
//  4) After a notify is issued, there is no guarantee that
//     upon release of the mutex that the next mutex lock
//     will go to a thread waiting at a condition wait.  In
//     particular, consider this scenario:
//
//        Initially, count=0, thread 1 has a lock on mutex.
//
//        Thread 1:
//           Tcl_ConditionWait(&cond,&mutex,NULL);
//           cout << count;
//           Tcl_MutexUnlock(mutex);
//
//        Threads 2 & 3:
//           Tcl_Lock(&mutex);
//           ++count;
//           Tcl_Notify(&cond);
//           Tcl_Unlock(&mutex);
//
//     The value for count output by thread 1 may be either 1 or 2.
//     For example, suppose thread 2 succeeds in acquire mutex first.
//     Then thread 2 increments count to 1, and notifies on cond.  Then
//     if thread 1 acquires mutex next, count will be 1.  However, it
//     is just as likely that thread 3 will acquire mutex, in which case
//     thread 3 will increment count to 2.  Assuming that no other
//     threads are waiting for mutex, then thread 1 will finally acquire
//     mutex, and the value for count at this time will be 2.
//        Similarly, even if there is only one thread waiting on a
//     particular condition, the number of time that condition is
//     notified may be greater than the number of times the waiting
//     thread is awaken.  For example, if Thread 1 above is changed to
//
//        Thread 1: (BAD BAD BAD!)
//           Tcl_ConditionWait(&cond,&mutex,NULL);
//           cout << count;
//           Tcl_ConditionWait(&cond,&mutex,NULL);
//           cout << count;
//           Tcl_MutexUnlock(mutex);
//
//     then this may hang at the second wait.  The correct way to
//     wait on multiple notifies is like so:
//
//        Thread 1: (GOOD)
//           while(count<2) { // Assume mutex held coming in
//              Tcl_ConditionWait(&cond,&mutex,NULL);
//           }
//           count << count;
//           Tcl_MutexUnlock(mutex);
//


// Thread-safe fprintf
void Oxs_ThreadPrintf(FILE* fptr,const char *format, ...) {
  static Tcl_Mutex oxs_threadprintf_mutex = 0;
  Tcl_MutexLock(&oxs_threadprintf_mutex);

  va_list arg_ptr;
  va_start(arg_ptr,format);
  vfprintf(fptr,format,arg_ptr);
  va_end(arg_ptr);

  fflush(fptr);

  Tcl_MutexUnlock(&oxs_threadprintf_mutex);
}

////////////////////////////////////////////////////////////////////////
// Mainline for actual (OS) threads
Tcl_ThreadCreateProc _Oxs_Thread_threadmain;
Tcl_ThreadCreateType _Oxs_Thread_threadmain(ClientData clientdata)
{ // NB: Be careful, this routine and any routine called from inside
  //     must be re-entrant!

  // Cast and dereference clientdata
  // NB: It is assumed that the referenced object remains valid
  // throughout the life of the thread.
  Oxs_Thread& othread = *static_cast<Oxs_Thread *>(clientdata);

  // Break out references to object data.  Copy thread_number on the
  // assumption that it is fixed throughout object lifetime.
  // The stop, runobj, and data objects may change between calls
  // (i.e., wakeups); we hold a references to these pointers to keep
  // this up-to-date.
  Oxs_ThreadControl& start  = othread.start;
  int thread_number         = othread.thread_number;
  Oxs_ThreadControl* &stop  = othread.stop;
  Oxs_ThreadRunObj* &runobj = othread.runobj;
  void* &data               = othread.data;

  // Notify parent Oxs_Thread that we are ready and waiting
  Tcl_MutexLock(&start.mutex);
  start.count = 1; // Ready to wait
  Tcl_ConditionNotify(&start.cond);

  // Action loop
  while(1) {
    while(start.count>0) {
      Tcl_ConditionWait(&start.cond, &start.mutex, 0);
    }
    if(0==runobj) break;
    runobj->Cmd(thread_number,data);
    Tcl_MutexLock(&stop->mutex);
    --stop->count; // NB: stop->mutex must be acquired before
                  ///     changing stop->count.
    Tcl_ConditionNotify(&stop->cond);
    start.count = 1; // Ready to wait.  Set this before freeing stop
    /// mutex, so that the thread signals as ready if tested from
    /// inside Oxs_ThreadTree::Launch.
    Tcl_MutexUnlock(&stop->mutex);
  }
  start.count = 1;
  Tcl_ConditionNotify(&start.cond);
  Tcl_MutexUnlock(&start.mutex);
  Tcl_FinalizeThread();
  TCL_THREAD_CREATE_RETURN;
}

////////////////////////////////////////////////////////////////////////
// Oxs_Thread class, which is control object for actual thread.
Oxs_Thread::Oxs_Thread
(int thread_number_x)
  : thread_number(thread_number_x),
    stop(0), runobj(0), data(0)
{
  Tcl_MutexLock(&start.mutex);
  if(TCL_OK
     != Tcl_CreateThread(&thread_id, _Oxs_Thread_threadmain, this,
                         TCL_THREAD_STACK_DEFAULT, TCL_THREAD_NOFLAGS)) {
    Tcl_MutexUnlock(&start.mutex);
    OXS_THROW(Oxs_BadResourceAlloc,"Thread creation failed.");
  }
  start.count = 0;
  while(0 == start.count) {
    // I don't know that this loop is really necessary, because
    // this is the only thread waiting on start.cond and only
    // one thread is notifying on start.cond, but all the
    // references say to do this, and it shouldn't hurt.
    Tcl_ConditionWait(&start.cond, &start.mutex, 0);
  }
  Tcl_MutexUnlock(&start.mutex);
}

Oxs_Thread::~Oxs_Thread()
{
  Tcl_MutexLock(&start.mutex);
  stop = 0;
  runobj = 0;
  data = 0;
  start.count = 0;
  Tcl_ConditionNotify(&start.cond);
  while(0 == start.count) {
    // Is this loop necessary?  Cf. notes in check_start
    // loop in Oxs_Thread constructor.
    Tcl_ConditionWait(&start.cond, &start.mutex, 0);
  }
  Tcl_MutexUnlock(&start.mutex);
  // Note: start.cond and start.mutex are automatically finalized
  // inside the Oxs_ThreadControl destructor
}

void Oxs_Thread::RunCmd
(Oxs_ThreadControl& stop_x,
 Oxs_ThreadRunObj& runobj_x,
 void* data_x)
{
  // runobj == NULL is used as a message to threadmain to
  // terminate.  Don't expose this implementation detail
  // to the user.  Instead, raise an error.
  if(0 == &runobj_x) {
    OXS_THROW(Oxs_BadParameter,"NULL thread runobj reference.");
  }
  Tcl_MutexLock(&start.mutex);
  stop = &stop_x;
  runobj = &runobj_x;
  data = data_x;
  start.count = 0; // Run signal
  Tcl_ConditionNotify(&start.cond);
  Tcl_MutexUnlock(&start.mutex);
}

////////////////////////////////////////////////////////////////////////
// ThreadTree class, manager for Oxs_Thread objects
// Note: stop.mutex is grabbed by Oxs_ThreadTree upon the
//       first Launch command, and is held until Join
//       is called.  Each tree has its own stop control,
//       but the thread array is shared by all trees.
Tcl_Mutex Oxs_ThreadTree::launch_mutex = 0;
std::vector<Oxs_Thread*> Oxs_ThreadTree::threads;

void Oxs_ThreadTree::Launch(Oxs_ThreadRunObj& runobj,void* data)
{ // Launched threads, one at a time.
  if(0==threads_unjoined) Tcl_MutexLock(&stop.mutex);

  Tcl_MutexLock(&launch_mutex); // Make sure no new threads get
  /// run while we're looking for a free thread.
  int free_thread=0;
  while(size_t(free_thread)<threads.size()) {
    if(!threads[free_thread]->IsRunning()) break;
    ++free_thread;
  }
  if(size_t(free_thread)==threads.size()) {
    // All threads in use; create a new thread
    threads.push_back(new Oxs_Thread(free_thread));
fprintf(stderr,"Starting thread #%d\n",free_thread); /**/
  }
  ++threads_unjoined;
  ++stop.count;
  threads[free_thread]->RunCmd(stop,runobj,data);
  Tcl_MutexUnlock(&launch_mutex);
}

void Oxs_ThreadTree::Join()
{ // Joins all running threads, and releases mutex_stop
  if(0==threads_unjoined) return; 

  // Note: If threads_unjoined>0, then we should be
  //       holding stop.mutex
  while(stop.count>0) {
    Tcl_ConditionWait(&stop.cond, &stop.mutex, 0);
  }
  threads_unjoined = 0;
  Tcl_MutexUnlock(&stop.mutex);
}

void Oxs_ThreadTree::EndThreads()
{ // Note: This is a static member function of Oxs_ThreadTree
  for(size_t i=0;i<threads.size();++i) {
    delete threads[i];
  }
  threads.clear();
  Tcl_MutexFinalize(&launch_mutex);
  launch_mutex = 0;
}

#endif // OOMMF_THREADS
