/* FILE: modllg_eulerevolve.cc                 -*-Mode: c++-*-
 *
 * Concrete evolver class, using simple forward Euler steps.
 * This version implements a modified LLG equation with |m|
 * control.
 *
 */

#include <float.h>

#include "nb.h"
#include "director.h"
#include "timedriver.h"
#include "simstate.h"
#include "modllg_eulerevolve.h"
#include "key.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_ModLLG_EulerEvolve);

/* End includes */

// Constructor
Oxs_ModLLG_EulerEvolve::Oxs_ModLLG_EulerEvolve(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_TimeEvolver(name,newdtr,argstr),
    energy_state_id(0),next_timestep(0.)
{
  // Process arguments
  min_timestep=GetRealInitValue("min_timestep",0.);
  max_timestep=GetRealInitValue("max_timestep",1e-10);
  if(max_timestep<=0.0) {
    char buf[4096];
    Oc_Snprintf(buf,sizeof(buf),
		"Invalid parameter value:"
		" Specified max time step is %g (should be >0.)",
		max_timestep);
    throw Oxs_Ext::Error(this,buf);
  }

  allowed_error_rate = GetRealInitValue("error_rate",-1);
  if(allowed_error_rate>0.0) {
    allowed_error_rate *= PI*1e9/180.; // Convert from deg/ns to rad/s
  }
  allowed_absolute_step_error
    = GetRealInitValue("absolute_step_error",0.2);
  if(allowed_absolute_step_error>0.0) {
    allowed_absolute_step_error *= PI/180.; // Convert from deg to rad
  }
  allowed_relative_step_error
    = GetRealInitValue("relative_step_error",0.2);

  step_headroom = GetRealInitValue("step_headroom",0.85);
  if(step_headroom<=0.) {
    throw Oxs_Ext::Error(this,"Invalid initialization detected:"
			 " step_headroom value must be bigger than 0.");
  }

  alpha = GetRealInitValue("alpha",0.5);

  // User may specify either gamma_G (Gilbert) or
  // gamma_LL (Landau-Lifshitz).  Code uses "gamma"
  // which is LL form.
  if(HasInitValue("gamma_G") && HasInitValue("gamma_LL")) {
    throw Oxs_Ext::Error(this,"Invalid Specify block; "
			 "both gamma_G and gamma_LL specified.");
  } else if(HasInitValue("gamma_G")) {
    gamma = GetRealInitValue("gamma_G")/(1+alpha*alpha);
  } else if(HasInitValue("gamma_LL")) {
    gamma = GetRealInitValue("gamma_LL");
  } else {
    gamma = 2.211e5/(1+alpha*alpha);
  }
  gamma = fabs(gamma); // Force positive

  do_precess = GetIntInitValue("do_precess",1);

  start_dm = GetRealInitValue("start_dm",0.01);
  start_dm *= PI/180.; // Convert from deg to rad

  // |m| adjustment parameters
  resize_norm = RT_INVALID;
  String resize_norm_string = GetStringInitValue("resize_norm","dmdt_ave");
  Oxs_ToLower(resize_norm_string);
  if(resize_norm_string.compare("dmdt_ave")==0) {
    resize_norm = RT_DMDT_AVE;
  } else if(resize_norm_string.compare("norm")==0) {
    resize_norm = RT_NORM;
  } else if(resize_norm_string.compare("none")==0) {
    resize_norm = RT_NONE;
  } else {
    resize_norm = RT_INVALID; // Safety
  }
  if(RT_INVALID == resize_norm) {
    throw Oxs_Ext::Error(this,"Invalid resize_norm setting detected."
			 " Should be: \"dmdt_ave\", \"norm\", or \"none\".");
  }
  allowed_mnorm_deviation = GetRealInitValue("norm_error",0.0);

  resize_differential = RT_INVALID;
  String resize_differential_string
    = GetStringInitValue("resize_differential","dmdt_ave");
  Oxs_ToLower(resize_differential_string);
  if(resize_differential_string.compare("dmdt_ave")==0) {
    resize_differential = RT_DMDT_AVE;
  } else if(resize_differential_string.compare("norm")==0) {
    resize_differential = RT_NORM;
  } else if(resize_differential_string.compare("none")==0) {
    resize_differential = RT_NONE;
  } else {
    resize_differential = RT_INVALID; // Safety
  }
  if(RT_INVALID == resize_differential) {
    throw Oxs_Ext::Error(this,"Invalid resize_differential setting detected."
			 " Should be: \"dmdt_ave\", \"norm\", or \"none\".");
  }
  m_mag_adj = GetRealInitValue("mag_adj",1.0);
  m_mag_ie_adj = GetRealInitValue("mag_ie_adj",1.0);
  REAL8m m_mag_ie_period = GetRealInitValue("mag_ie_period",0.1e-9);

  if(m_mag_ie_period<=0.0) {
    throw Oxs_Ext::Error(this,"Invalid mag_ie_period specified."
			 " Should be integration period, in ns; must be >0.");
  }
  m_mag_ie_freq = 1.0/m_mag_ie_period;

  m_mag_adj *= 1.0;       // Default scaling
#if 0
  m_mag_ie_adj *= 1.0;    // Default scaling
#else
  m_mag_ie_adj *= m_mag_ie_freq;    // Default scaling
#endif

  // Setup outputs
  max_dm_dt_output.Setup(this,InstanceName(),"Max dm/dt","deg/ns",0,
     &Oxs_ModLLG_EulerEvolve::UpdateDerivedOutputs);
  min_m_output.Setup(this,InstanceName(),"Min |m|","",0,
     &Oxs_ModLLG_EulerEvolve::UpdateDerivedOutputs);
  max_m_output.Setup(this,InstanceName(),"Max |m|","",0,
     &Oxs_ModLLG_EulerEvolve::UpdateDerivedOutputs);
  max_mdev_output.Setup(this,InstanceName(),"Max ||m|-1|","",0,
     &Oxs_ModLLG_EulerEvolve::UpdateDerivedOutputs);
  dE_dt_output.Setup(this,InstanceName(),"dE/dt","J/s",0,
     &Oxs_ModLLG_EulerEvolve::UpdateDerivedOutputs);
  delta_E_output.Setup(this,InstanceName(),"Delta E","J",0,
     &Oxs_ModLLG_EulerEvolve::UpdateDerivedOutputs);
  dm_dt_output.Setup(this,InstanceName(),"dm/dt","rad/s",1,
     &Oxs_ModLLG_EulerEvolve::UpdateDerivedOutputs);
  mxH_output.Setup(this,InstanceName(),"mxH","A/m",1,
     &Oxs_ModLLG_EulerEvolve::UpdateDerivedOutputs);

  max_dm_dt_output.Register(director,-5);
  min_m_output.Register(director,-5);
  max_m_output.Register(director,-5);
  max_mdev_output.Register(director,-5);
  dE_dt_output.Register(director,-5);
  delta_E_output.Register(director,-5);
  dm_dt_output.Register(director,-5);
  mxH_output.Register(director,-5);

  // dm_dt and mxH output caches are used for intermediate storage,
  // so enable caching.
  dm_dt_output.CacheRequestIncrement(1);
  mxH_output.CacheRequestIncrement(1);

  VerifyAllInitArgsUsed();
}

BOOL Oxs_ModLLG_EulerEvolve::Init()
{
  energy_state_id=0;   // Mark as invalid state
  next_timestep=0.;    // Dummy value

  integrated_norm_error = 0.0;      // Ensure integrated error is reset.
  new_integrated_norm_error = 0.0;

  return Oxs_TimeEvolver::Init();
}


Oxs_ModLLG_EulerEvolve::~Oxs_ModLLG_EulerEvolve()
{}


void Oxs_ModLLG_EulerEvolve::Calculate_dm_dt
(const Oxs_Mesh& mesh_,
 const Oxs_MeshValue<REAL8m>& Ms_,
 const Oxs_MeshValue<ThreeVector>& mxH_,
 const Oxs_MeshValue<ThreeVector>* old_dm_dt_ptr_,
 const Oxs_MeshValue<ThreeVector>& spin_,
 REAL8m pE_pt_, REAL8m timestep_,
 Oxs_MeshValue<ThreeVector>& dm_dt_,
 REAL8m& max_dm_dt_,REAL8m& dE_dt_,REAL8m& min_timestep_)
{ // Imports: mesh_, Ms_, mxH_, H_, spin_, pE_pt_
  // Exports: dm_dt_, max_dm_dt_, dE_dt_, min_timestep_
  //          min_m_, max_m_

  const UINT4m size = mesh_.Size(); // Assume all imports are compatible
  REAL8m coef1 = -fabs(gamma);
  REAL8m coef2 = -fabs(alpha);
  ThreeVector scratch;
  dm_dt_.AdjustSize(&mesh_);

  if(!integrated_norm_error.CheckMesh(&mesh_)) {
    integrated_norm_error.AdjustSize(&mesh_);
    integrated_norm_error = 0.0;
  }
  if(!new_integrated_norm_error.CheckMesh(&mesh_)) {
    new_integrated_norm_error.AdjustSize(&mesh_);
    new_integrated_norm_error = 0.0;
  }

  UINT4m i;
  REAL8m ie_step_exp = exp(-1*m_mag_ie_freq*timestep_);
  for(i=0;i<size;i++) {
    if(Ms_[i]==0) {
      dm_dt_[i].Set(0.0,0.0,0.0);
    } else {
      scratch = mxH_[i];
      scratch *= coef1;
      if(do_precess) {
	dm_dt_[i] = scratch;
      } else {
	dm_dt_[i].Set(0.0,0.0,0.0);
      }

      REAL8m m_mag = sqrt(spin_[i].MagSq());

      scratch ^= spin_[i]; // -|gamma|((mxH)xm)
      scratch *= coef2;  // |alpha.gamma|((mxH)xm) = -|alpha.gamma|(mx(mxH))
      dm_dt_[i] += scratch;

      // Add |m| adjustment
      if(RT_NONE != resize_differential) {
        REAL8m error = m_mag - 1.0;
        if(RT_DMDT_AVE==resize_differential) {
          if(NULL == old_dm_dt_ptr_ || 0.0 == timestep_) {
            scratch.Set(0.,0.,0.); // We can't compute rotation
            /// axis, so make no correction.
          } else {
            ThreeVector utmp = (*old_dm_dt_ptr_)[i];
            REAL8m odmdtsq = utmp.MagSq();
            utmp ^= dm_dt_[i];
            REAL8m ndmdtsq = dm_dt_[i].MagSq();
            REAL8m umagsq = utmp.MagSq();
            if(umagsq <= odmdtsq*ndmdtsq*1e-16) {
              // dm_dt's too close to extract rotation axis; punt
              scratch = spin_[i];
              scratch.MakeUnit();
            } else {
              utmp *= 1.0/sqrt(umagsq);
              REAL8m h = fabs(spin_[i] * utmp);
              if(h>=1.0-REAL8_EPSILON) {
                // Projection falls outside sphere; punt
                scratch = spin_[i];
                scratch.MakeUnit();
              } else {
                scratch = utmp;       // Should this be u x spin x u
                scratch ^= spin_[i];  // or u x dm_dt ???
                scratch ^= utmp;
                scratch.MakeUnit();
                error = scratch*spin_[i]/sqrt(1-h*h) - 1.0;
              }
            }
          }
        } else {
          // resize_differential == RT_NORM
          scratch = spin_[i];
          if(m_mag>=0.5 || 1.0<m_mag*DBL_MAX) scratch *= 1.0/m_mag;
        }

#if 0
        REAL8m norm_error = 1.0 - m_mag;
        REAL8m proportional_scale = m_mag_adj * norm_error;
        REAL8m inm = ie_step_exp*integrated_norm_error[i]
          + timestep_*norm_error;

        new_integrated_norm_error[i] = inm;
        REAL8m ie_scale = m_mag_ie_adj*inm;

        if(timestep_>1.0
           || fabs(proportional_scale) < (0.125*DBL_MAX)*timestep_ ) {
          proportional_scale /= timestep_;
        } else {
          proportional_scale = 0.0; // Safety
        }
        if(timestep_>1.0
           || fabs(ie_scale) < (0.125*DBL_MAX)*timestep_*timestep_ ) {
          ie_scale /= (timestep_*timestep_);
        } else {
          ie_scale = 0.0; // Safety
        }
        scratch *= (proportional_scale + ie_scale);
#else

#if 0
        REAL8m inm = ie_step_exp*integrated_norm_error[i]
          + timestep_*error;
        new_integrated_norm_error[i] = inm;

        REAL8m sign = -1.0;
        if(error<0.0) {
          error *= -1;
          sign = 1.0;
        }
        REAL8m scale = sign*sqrt(dm_dt_[i].MagSq())
          *sqrt((2+error)*error)
          *(m_mag_adj + m_mag_ie_adj*inm);
#else
        error += m_mag_ie_adj*integrated_norm_error[i];

        REAL8m inm = ie_step_exp*integrated_norm_error[i]
          + timestep_*error;
        new_integrated_norm_error[i] = inm;

        REAL8m sign = -1.0;
        if(error<0.0) {
          error *= -1;
          sign = 1.0;
        }
        REAL8m scale = sign*sqrt(dm_dt_[i].MagSq())
          *sqrt((2+error)*error)*m_mag_adj;
#endif

#if 0
fprintf(stderr,"error=%g\n..scale=%g\n..adj=%g, ie_adj=%g\n"
        "..inm=%g\n",error*sign*-1,scale,m_mag_adj,m_mag_ie_adj,inm); /**/
#endif

        scratch *= scale;
#endif

        dm_dt_[i] += scratch; 
      }

    }
  }

  // Zero dm_dt at fixed spin sites
  UpdateFixedSpinList(&mesh_);
  const UINT4m fixed_count = GetFixedSpinCount();
  for(UINT4m j=0;j<fixed_count;j++) {
    dm_dt_[GetFixedSpin(j)].Set(0.,0.,0.); // What about pE_pt???
  }

  // Collect statistics
  REAL8m max_dm_dt_sq=0.0;
  REAL8m dE_dt_sum=0.0;
  UINT4m max_index=0;
  for(i=0;i<size;i++) {
    REAL8m dm_dt_sq = dm_dt_[i].MagSq();
    if(dm_dt_sq>0.0) {
      dE_dt_sum += mxH_[i].MagSq() * Ms_[i] * mesh_.Volume(i);
      /// asdfasdf /**/ Is this calculus correct in the modllg case???
      /// Answer: NO! for norm differential case.
      ///         Approximately correct for dm/dt1 x dm/dt2,
      ///         if H is approximately parallel to dm/dt1 x dm/dt2.
      if(dm_dt_sq>max_dm_dt_sq) {
	max_dm_dt_sq=dm_dt_sq;
	max_index = i;
      }
    }
  }
  max_dm_dt_ = sqrt(max_dm_dt_sq);
  dE_dt_ = -1 * MU0 * fabs(gamma*alpha) * dE_dt_sum + pE_pt_;
  /// The first term is (partial E/partial M)*dM/dt, the
  /// second term is (partial E/partial t)*dt/dt.  Note that,
  /// provided Ms_[i]>=0, that by constructions dE_dt_sum above
  /// is always non-negative, so dE_dt_ can only be made positive
  /// by positive pE_pt_.

  // Get bound on smallest stepsize that would actually
  // change spin new_max_dm_dt_index:
  REAL8m min_ratio = DBL_MAX/2.;
  if(fabs(dm_dt_[max_index].x)>=1.0 ||
     min_ratio*fabs(dm_dt_[max_index].x) > fabs(spin_[max_index].x)) {
    min_ratio = fabs(spin_[max_index].x/dm_dt_[max_index].x);
  }
  if(fabs(dm_dt_[max_index].y)>=1.0 ||
     min_ratio*fabs(dm_dt_[max_index].y) > fabs(spin_[max_index].y)) {
    min_ratio = fabs(spin_[max_index].y/dm_dt_[max_index].y);
  }
  if(fabs(dm_dt_[max_index].z)>=1.0 ||
     min_ratio*fabs(dm_dt_[max_index].z) > fabs(spin_[max_index].z)) {
    min_ratio = fabs(spin_[max_index].z/dm_dt_[max_index].z);
  }
  min_timestep_ = min_ratio * REAL8_EPSILON;

  // A timestep of size min_timestep will be hopelessly lost
  // in roundoff error.  So increase a bit, based on an empirical
  // fudge factor.  This fudge factor can be tested by running a
  // problem with start_dm = 0.  If the evolver can't climb it's
  // way out of the stepsize=0 hole, then this fudge factor is too
  // small.  So far, the most challenging examples have been
  // problems with small cells with nearly aligned spins, e.g. in
  // a remanent state, where an external field is applied at t=0.
  // Damping ratio doesn't seem to have much effect, either way.
  min_timestep_ *= 8;

  return;
}


BOOL
Oxs_ModLLG_EulerEvolve::Step(const Oxs_TimeDriver* driver,
		      Oxs_ConstKey<Oxs_SimState> current_state,
                      const Oxs_DriverStepInfo& /* step_info */,
		      Oxs_Key<Oxs_SimState>& next_state)
{
  const REAL8m max_step_increase = 1.25;
  const REAL8m max_step_decrease = 0.5;

  UINT4m size,i; // Mesh size and indexing variable

  const Oxs_SimState& cstate = current_state.GetReadReference();

  // Do first part of next_state structure initialization.
  Oxs_SimState& workstate = next_state.GetWriteReference();
  driver->FillState(current_state.GetReadReference(),workstate);

  if(cstate.mesh->Id() != workstate.mesh->Id()) {
    throw Oxs_Ext::Error(this,
       "Oxs_ModLLG_EulerEvolve::Step: Oxs_Mesh not fixed across steps.");
  }

  if(cstate.Id() != workstate.previous_state_id) {
    throw Oxs_Ext::Error(this,
       "Oxs_ModLLG_EulerEvolve::Step: State continuity break detected.");
  }

  // Pull cached values out from cstate.
  // If cstate.Id() == energy_state_id, then cstate has been run
  // through either this method or UpdateDerivedOutputs.  Either
  // way, all derived state data should be stored in cstate,
  // except currently the "energy" mesh value array, which is
  // stored independently inside *this.  Eventually that should
  // probably be moved in some fashion into cstate too.
  if(energy_state_id != cstate.Id()) {
    // cached data out-of-date
    UpdateDerivedOutputs(cstate);
  }
  BOOL cache_good = 1;
  REAL8m max_dm_dt,dE_dt,delta_E,pE_pt;
  REAL8m timestep_lower_bound;  // Smallest timestep that can actually
  /// change spin with max_dm_dt (due to REAL8_EPSILON restrictions).
  /// The next timestep is based on the error from the last step.  If
  /// there is no last step (either because this is the first step,
  /// or because the last state handled by this routine is different
  /// from the incoming current_state), then timestep is calculated
  /// so that max_dm_dt * timestep = start_dm.

  cache_good &= cstate.GetDerivedData("Max dm/dt",max_dm_dt);
  cache_good &= cstate.GetDerivedData("dE/dt",dE_dt);
  cache_good &= cstate.GetDerivedData("Delta E",delta_E);
  cache_good &= cstate.GetDerivedData("pE/pt",pE_pt);
  cache_good &= cstate.GetDerivedData("Timestep lower bound",
				      timestep_lower_bound);
  cache_good &= (energy_state_id == cstate.Id());
  cache_good &= (dm_dt_output.cache.state_id == cstate.Id());

  if(!cache_good) {
    throw Oxs_Ext::Error(this,
       "Oxs_ModLLG_EulerEvolve::Step: Invalid data cache.");
  }

  const Oxs_MeshValue<ThreeVector>& dm_dt = dm_dt_output.cache.value;

  // Negotiate with driver over size of next step.
  REAL8m stepsize = next_timestep;
  if(stepsize<=0.0) {
    if(max_dm_dt>=1.0 || start_dm  < max_timestep * max_dm_dt) {
      stepsize = start_dm / max_dm_dt;
    } else {
      stepsize = max_timestep;
    }
  }
  if(stepsize>max_timestep) stepsize = max_timestep;
  if(stepsize<min_timestep) stepsize = min_timestep;

  workstate.last_timestep=stepsize;
  if(stepsize<timestep_lower_bound) {
    workstate.last_timestep=timestep_lower_bound;
  }

  if(cstate.stage_number != workstate.stage_number) {
    // New stage
    // Should we reset integrated_error cache across stage
    // boundaries?  Currently we choose not to, but in any
    // event we can't change it here because in practice
    // THIS CODE BRANCH IS NOT ACTIVE!  At present (Oct 2005),
    // stage boundaries are handled by the driver, which calls
    // max_dm_dt_obj_ptr->Output (here UpdateDerivedOutputs)
    // from Oxs_TimeDriver::ChildIsStageDone rather than Step.
    workstate.stage_start_time = cstate.stage_start_time
                                + cstate.stage_elapsed_time;
    workstate.stage_elapsed_time = workstate.last_timestep;
  } else {
    workstate.stage_start_time = cstate.stage_start_time;
    workstate.stage_elapsed_time = cstate.stage_elapsed_time
                                  + workstate.last_timestep;
  }
  workstate.iteration_count = cstate.iteration_count + 1;
  workstate.stage_iteration_count = cstate.stage_iteration_count + 1;

  driver->FillStateSupplemental(workstate);
  BOOL forcestep=0;
  if(workstate.last_timestep>stepsize ||
     workstate.last_timestep<=min_timestep) {
    // Either driver wants to force this stepsize,
    // or else suggested stepsize is smaller than
    // timestep_lower_bound and/or min_timestep.
    forcestep=1;
  }
  stepsize = workstate.last_timestep;

  // Put new spin configuration in next_state
  workstate.spin.AdjustSize(workstate.mesh); // Safety
  size = workstate.spin.Size();
  ThreeVector tempspin;
  REAL8m min_m = 1.0;
  REAL8m max_m = 1.0;
  for(i=0;i<size;++i) {
    tempspin = dm_dt[i];
    tempspin *= stepsize;
    tempspin += cstate.spin[i];
    REAL8m normsq = tempspin.MagSq();
    if(normsq<min_m) min_m = normsq;
    if(normsq>max_m) max_m = normsq;
    workstate.spin[i] = tempspin;
  }
  min_m = sqrt(min_m);
  max_m = sqrt(max_m);
  REAL8m max_norm_dev = 1.0 - min_m;
  if(max_norm_dev < max_m - 1.0) max_norm_dev = max_m - 1.0;

  // For resize_norm ==  RT_H, delay size adjustment until after
  // new dm_dt is computed.
  if( RT_NORM == resize_norm && max_norm_dev > allowed_mnorm_deviation) {
    min_m = max_m = 1.0;
    for(i=0;i<size;++i) {
      workstate.spin[i].MakeUnit();
      REAL8m normsq = workstate.spin[i].MagSq();
      if(normsq<min_m) min_m = normsq;
      if(normsq>max_m) max_m = normsq;
    }
    min_m = sqrt(min_m);
    max_m = sqrt(max_m);
    max_norm_dev = 1.0 - min_m;
    if(max_norm_dev < max_m - 1.0) max_norm_dev = max_m - 1.0;
  }

  const Oxs_SimState& nstate
    = next_state.GetReadReference();  // Release write lock

  //  Calculate delta E
  REAL8m new_pE_pt;
  GetEnergyDensity(nstate,new_energy,
		   &mxH_output.cache.value,
		   NULL,new_pE_pt);
  mxH_output.cache.state_id=nstate.Id();
  const Oxs_MeshValue<ThreeVector>& mxH = mxH_output.cache.value;

  REAL8m dE=0.0;
  REAL8m var_dE=0.0;
  REAL8m total_E=0.0;
  for(i=0;i<size;++i) {
    REAL8m vol = nstate.mesh->Volume(i);
    REAL8m e = energy[i];
    total_E += e * vol;
    REAL8m new_e = new_energy[i];
    dE += (new_e - e) * vol;
    var_dE += (new_e*new_e + e*e)*vol*vol;
  }
  var_dE *= 256*REAL8_EPSILON*REAL8_EPSILON; // Variance, assuming
  /// error in each energy[i] term is independent, uniformly
  /// distributed, 0-mean, with range +/- 16*REAL8_EPSILON*energy[i].
  /// It would probably be better to get an error estimate directly
  /// from each energy term.

  // Get error estimate.  See step size adjustment discussion in
  // MJD Notes II, p72 (18-Jan-2001).

  REAL8m new_max_dm_dt,new_dE_dt,new_timestep_lower_bound;
  Calculate_dm_dt(*(nstate.mesh),*(nstate.Ms),mxH,&dm_dt,
                  nstate.spin,new_pE_pt,stepsize,
                  new_dm_dt,new_max_dm_dt,new_dE_dt,
                  new_timestep_lower_bound);

  REAL8m max_error=0;
  for(i=0;i<size;++i) {
    ThreeVector temp = dm_dt[i];
    temp -= new_dm_dt[i];
    REAL8m temp_error = temp.MagSq();
    if(temp_error>max_error) max_error = temp_error;
  }
  max_error = sqrt(max_error)/2.0; // Actual (local) error
  /// estimate is max_error * stepsize

  // Energy check control
#ifdef FOO
  REAL8m expected_dE = 0.5 * (dE_dt+new_dE_dt) * stepsize;
  REAL8m dE_error = dE - expected_dE;
  REAL8m max_allowed_dE = expected_dE + 0.25*fabs(expected_dE);
  max_allowed_dE += REAL8_EPSILON*fabs(total_E);
  max_allowed_dE += 2*sqrt(var_dE);
#else
  REAL8m max_allowed_dE = 0.5 * (pE_pt+new_pE_pt) * stepsize
    + OC_MAX(REAL8_EPSILON*fabs(total_E),2*sqrt(var_dE));
  /// The above says essentially that the spin adjustment can
  /// increase the energy by only as much as pE/pt allows; in
  /// the absence of pE/pt, the energy should decrease.  I
  /// think this may be problematic, if at the start of a stage
  /// the spins are near equilibrium, and the applied field is
  /// ramping up slowly.  In this case there won't be much "give"
  /// in the spin configuration with respect to pE/pm.  But I
  /// haven't seen an example of this yet, so we'll wait and see.
  /// -mjd, 27-July-2001.
#endif

  // Check step and adjust next_timestep.  The relative error
  // check is a bit fudged, because rather than limiting the
  // relative error uniformly across the sample, we limit it
  // only at the position that has the maximum absolute error
  // (i.e., max_error is max *absolute* error).  I haven't
  // tested to see if uniformly limiting relative error is
  // workable (it might be too restrictive for most purposes),
  // but the present setup seems to solve the problem of convergence
  // near equilibrium.  -mjd, 2001-02-23.
  // NOTE: Since all three error controls (error_rate,
  //  absolute_step_error, and relative_step_error) assume error
  //  grows linearly with step size, we can check up front to see
  //  which control is most restrictive, store that constraint in
  //  working_allowed_error, and then adjust the step size without
  //  regard to which control is being exercised.
  REAL8m working_allowed_error
    = max_step_increase*max_error/step_headroom;
  if(allowed_error_rate>=0.
     && working_allowed_error>allowed_error_rate) {
    working_allowed_error=allowed_error_rate;
  }
  if(allowed_absolute_step_error>=0.
     && stepsize*working_allowed_error>allowed_absolute_step_error) {
    working_allowed_error=allowed_absolute_step_error/stepsize;
  }
  if(allowed_relative_step_error>=0.
     && working_allowed_error>allowed_relative_step_error*max_dm_dt) {
    working_allowed_error = allowed_relative_step_error * max_dm_dt;
  }
  if(!forcestep) {
    next_timestep=1.0;  // Size relative to current step
    if(max_error>working_allowed_error) {
      next_timestep = step_headroom*working_allowed_error/max_error;
    } else if(dE>max_allowed_dE) {
      // Energy check
      next_timestep=0.5;
    }
    if(next_timestep<1.0) {
      // Reject step
      if(next_timestep<max_step_decrease)
	next_timestep=max_step_decrease;
      next_timestep *= stepsize;
      return 0;
    }
  }

  // Otherwise, accept step.

  // Calculate next step using estimate of step size that
  // would just meet the error restriction (with "headroom"
  // safety margin).
  next_timestep = max_step_increase;
  if(next_timestep*max_error>step_headroom*working_allowed_error) {
    next_timestep = step_headroom*working_allowed_error/max_error;
  }
  if(next_timestep<max_step_decrease)
    next_timestep=max_step_decrease;
  next_timestep *= stepsize;


  // If next_timestep is much smaller than the timestep
  // that brought us to cstate, then the reduction was
  // probably due to a stage stopping_time requirement.
  // So in this case bump the next_stepsize request up
  // to the minimum allowed reduction.
  REAL8m timestep_lower_limit = max_step_decrease * step_headroom;
  timestep_lower_limit *= timestep_lower_limit * cstate.last_timestep;
  if(next_timestep<timestep_lower_limit) {
    next_timestep = timestep_lower_limit;
  }

  if( RT_DMDT_AVE == resize_norm
      && max_norm_dev > allowed_mnorm_deviation) {
    // Cheat: Although nstate is locked down, directly vary the nstate
    // spin array to project m's back to the unit sphere.
    Oxs_MeshValue<ThreeVector>& tspin
      = const_cast<Oxs_MeshValue<ThreeVector>&>(nstate.spin);
    REAL8m min_normsq = DBL_MAX;
    REAL8m max_normsq = 0.0;
    for(i=0;i<size;++i) {
      // First find unit vector of approximate rotation axis,
      // via dm_dt_old x dm_dt_new
      ThreeVector utmp = dm_dt[i];
      REAL8m cdmdtsq = dm_dt[i].MagSq();
      utmp ^= new_dm_dt[i];
      REAL8m ndmdtsq = new_dm_dt[i].MagSq();
      REAL8m umagsq = utmp.MagSq();
      if(umagsq <= cdmdtsq*ndmdtsq*1e-16) {
        // dm_dt's too close to extract rotation axis; punt.
        tspin[i].MakeUnit();
      } else {
        utmp *= (1.0/sqrt(umagsq));
        ThreeVector vtmp = utmp;
        vtmp ^= tspin[i];
        REAL8m Ecos = tspin[i]*utmp;
        REAL8m Esin = sqrt(vtmp.MagSq());
        if(fabs(Ecos)>=1.0 || (fabs(Esin)<1.0 && 1.0>DBL_MAX*fabs(Esin))) {
          // Bad spin; just renormalize
          tspin[i].MakeUnit();
        } else {
          vtmp ^= utmp;
          vtmp *= (1.0/Esin);
          REAL8m B = sqrt(1-Ecos*Ecos) - Esin;
          vtmp *= B;
          tspin[i] += vtmp;
        }
      }
      REAL8m normsq = tspin[i].MagSq();
      if(normsq<min_normsq) min_normsq=normsq;
      if(normsq>max_normsq) max_normsq=normsq;
    }
    min_m = sqrt(min_normsq);
    max_m = sqrt(max_normsq);
    max_norm_dev = 1.0 - min_m;
    if(max_norm_dev < max_m - 1.0) max_norm_dev = max_m - 1.0;
  }

  if(!nstate.AddDerivedData("Timestep lower bound",
			    new_timestep_lower_bound) ||
     !nstate.AddDerivedData("Max dm/dt",new_max_dm_dt) ||
     !nstate.AddDerivedData("Min |m|",min_m) ||
     !nstate.AddDerivedData("Max |m|",max_m) ||
     !nstate.AddDerivedData("Max ||m|-1|",max_norm_dev) ||
     !nstate.AddDerivedData("dE/dt",new_dE_dt) ||
     !nstate.AddDerivedData("Delta E",dE) ||
     !nstate.AddDerivedData("pE/pt",new_pE_pt)) {
    throw Oxs_Ext::Error(this,
       "Oxs_ModLLG_EulerEvolve::Step:"
       " Programming error; data cache already set.");
  }

  dm_dt_output.cache.value.Swap(new_dm_dt);
  dm_dt_output.cache.state_id = nstate.Id();

  energy_state_id = nstate.Id();
  energy.Swap(new_energy);

  if(RT_NONE != resize_differential) {
    integrated_norm_error.Swap(new_integrated_norm_error);
  }

  return 1;  // Good step
}

void Oxs_ModLLG_EulerEvolve::UpdateDerivedOutputs(const Oxs_SimState& state)
{ // This routine fills all the Oxs_ModLLG_EulerEvolve Oxs_ScalarOutput's
  // to the appropriate value based on the import "state", and any of
  // Oxs_VectorOutput's that have CacheRequest enabled are filled.
  // It also makes sure all the expected WOO objects in state are
  // filled.
  max_dm_dt_output.cache.state_id
    = min_m_output.cache.state_id
    = max_m_output.cache.state_id
    = max_mdev_output.cache.state_id
    = dE_dt_output.cache.state_id
    = delta_E_output.cache.state_id
    = 0;  // Mark change in progress

  REAL8m dummy_value;
  if(!state.GetDerivedData("Max dm/dt",max_dm_dt_output.cache.value) ||
     !state.GetDerivedData("Min |m|",min_m_output.cache.value) ||
     !state.GetDerivedData("Max |m|",max_m_output.cache.value) ||
     !state.GetDerivedData("Max ||m|-1|",max_mdev_output.cache.value) ||
     !state.GetDerivedData("dE/dt",dE_dt_output.cache.value) ||
     !state.GetDerivedData("Delta E",delta_E_output.cache.value) ||
     !state.GetDerivedData("pE/pt",dummy_value) ||
     !state.GetDerivedData("Timestep lower bound",dummy_value) ||
     (dm_dt_output.GetCacheRequestCount()>0
      && dm_dt_output.cache.state_id != state.Id()) ||
     (mxH_output.GetCacheRequestCount()>0
      && mxH_output.cache.state_id != state.Id())) {

    // Missing at least some data, so calculate from scratch

    // Calculate H and mxH outputs
    Oxs_MeshValue<ThreeVector>& mxH = mxH_output.cache.value;

    Oxs_MeshValue<ThreeVector>* old_dm_dt = NULL;
    if(energy_state_id != 0 &&
       energy_state_id == dm_dt_output.cache.state_id) {
      old_dm_dt = &dm_dt_output.cache.value;
    }
    REAL8m pE_pt;
    GetEnergyDensity(state,energy,&mxH,NULL,pE_pt);
    energy_state_id=state.Id();
    mxH_output.cache.state_id=state.Id();
    if(!state.GetDerivedData("pE/pt",dummy_value)) {
      state.AddDerivedData("pE/pt",pE_pt);
    }

    // Calculate dm/dt, Max dm/dt and dE/dt
    Oxs_MeshValue<ThreeVector>& dm_dt
      = dm_dt_output.cache.value;
    dm_dt_output.cache.state_id=0;

    REAL8m timestep_lower_bound;
    Calculate_dm_dt(*(state.mesh),*(state.Ms),mxH,old_dm_dt,
                    state.spin,pE_pt,0.0,dm_dt,
		    max_dm_dt_output.cache.value,
		    dE_dt_output.cache.value,timestep_lower_bound);
    dm_dt_output.cache.state_id=state.Id();
    if(!state.GetDerivedData("Max dm/dt",dummy_value)) {
      state.AddDerivedData("Max dm/dt",max_dm_dt_output.cache.value);
    }

    const UINT4m size = state.mesh->Size();
    UINT4m i;
    REAL8m min_m=1.0;
    REAL8m max_m=1.0;
    for(i=0;i<size;++i) {
      REAL8m normsq = state.spin[i].MagSq();
      if(normsq<min_m) min_m = normsq;
      if(normsq>max_m) max_m = normsq;
    }
    min_m_output.cache.value = sqrt(min_m);
    max_m_output.cache.value = sqrt(max_m);
    if(!state.GetDerivedData("Min |m|",dummy_value)) {
      state.AddDerivedData("Min |m|",min_m_output.cache.value);
    }
    if(!state.GetDerivedData("Max |m|",dummy_value)) {
      state.AddDerivedData("Max |m|",max_m_output.cache.value);
    }
    if(!state.GetDerivedData("Max ||m|-1|",dummy_value)) {
      REAL8m lodev = 1.0 - min_m_output.cache.value;
      REAL8m hidev = max_m_output.cache.value - 1.0;
      state.AddDerivedData("Max ||m|-1|",(lodev>hidev?lodev:hidev));
    }
    if(!state.GetDerivedData("dE/dt",dummy_value)) {
      state.AddDerivedData("dE/dt",dE_dt_output.cache.value);
    }
    if(!state.GetDerivedData("Timestep lower bound",dummy_value)) {
      state.AddDerivedData("Timestep lower bound",
			   timestep_lower_bound);
    }

    if(!state.GetDerivedData("Delta E",dummy_value)) {
      if(state.previous_state_id!=0 && state.stage_iteration_count>0) {
	// Strictly speaking, we should be able to create dE for
	// stage_iteration_count==0 for stages>0, but as a practical
	// matter we can't at present.  Should give this more thought.
	// -mjd, 27-July-2001
	throw Oxs_Ext::Error(this,
	   "Oxs_ModLLG_EulerEvolve::UpdateDerivedOutputs:"
	   " Can't derive Delta E from single state.");
      }
      state.AddDerivedData("Delta E",0.0);
      dummy_value = 0.;
    }
    delta_E_output.cache.value=dummy_value;
  }

  max_dm_dt_output.cache.value*=(180e-9/PI);
  /// Convert from radians/second to deg/ns

  max_dm_dt_output.cache.state_id
    = min_m_output.cache.state_id
    = max_m_output.cache.state_id
    = max_mdev_output.cache.state_id
    = dE_dt_output.cache.state_id
    = delta_E_output.cache.state_id
    = state.Id();
}
